"""
Additional functionality that does not belong to the core of the Liesel framework.

While we consider functionality in this module to be reliably functional, we may choose
to organize it differently in the future. For example, submodules in ``liesel.contrib``
might be moved into their own, separate libraries at some point in the future.

The API in ``liesel.contrib`` is considered less stable than the API of the core Liesel
modules ``liesel.model`` and ``liesel.goose``.
"""
